#include <dlfcn.h>
#include <cstring>
#include <iostream>
#include <cstdlib>
#include "pkcs11_unix.h"
#include "base64.h"
using namespace std;

const string str_to_encrypt = "Text to encrypt";
const int AES_KEYLENGTH = 128;  // change this as desired for the security level you want

CK_FUNCTION_LIST_PTR initialize() {
    CK_FUNCTION_LIST_PTR p11;
    void *pDynLib;

    pDynLib = dlopen("/opt/fortanix/pkcs11/sdkms-pkcs11.so", RTLD_NOW);
    if (!pDynLib) {
        cout << "Failed to load the PKCS#11 library" << endl;
        return NULL;
    }
    cout << "Successfully initialized PKCS#11 library to use Fortanix SDKMS" << endl;
    CK_C_GetFunctionList pGetFunctionList = (CK_C_GetFunctionList) dlsym(pDynLib, "C_GetFunctionList");
    (*pGetFunctionList)(&p11);
    return p11;
}

CK_RV login(CK_FUNCTION_LIST_PTR p11, CK_SESSION_HANDLE_PTR phSession, string pin) {
    CK_RV rv;
    CK_SLOT_ID slotId = 1;

    rv = p11->C_Initialize(NULL_PTR);
    if (rv == CKR_OK) {
        rv = p11->C_OpenSession(slotId, CKF_SERIAL_SESSION | CKF_RW_SESSION, NULL_PTR, NULL_PTR, phSession);
        if (rv == CKR_OK) rv = p11->C_Login(*phSession, CKU_USER, (CK_UTF8CHAR_PTR)pin.c_str(), pin.length());
    }
    if (rv != CKR_OK) {
        cout << "Incorrect API Key. Error code = " << rv << endl;;
    } else {
        cout << "Successfully logged in to Fortanix SDKMS" << endl << endl;
    }
    return rv;
}

CK_OBJECT_HANDLE generate_key(CK_FUNCTION_LIST_PTR p11, CK_SESSION_HANDLE hSession, CK_ULONG len) {
    CK_RV rv;
    CK_MECHANISM mechKeyGen = {
      CKM_AES_KEY_GEN, NULL_PTR, 0
    };
    CK_BBOOL _true = CK_TRUE;
    CK_OBJECT_HANDLE hKey;

    CK_ATTRIBUTE keyTemplate[] = {
      {CKA_VALUE_LEN, &len, sizeof(len)},
      {CKA_ENCRYPT, &_true, sizeof(_true)},
      {CKA_DECRYPT, &_true, sizeof(_true)}
    };

    rv = p11->C_GenerateKey(hSession, &mechKeyGen, keyTemplate, sizeof(keyTemplate)/sizeof(*keyTemplate), &hKey);
    if (rv == CKR_OK) return hKey; else return CK_INVALID_HANDLE;
}

string encrypt(CK_FUNCTION_LIST_PTR p11, CK_SESSION_HANDLE hSession, CK_OBJECT_HANDLE hKey, string plain) {
    CK_RV rv;
    CK_BYTE *iv;
    CK_ULONG iv_len;
    string iv_b64;
    CK_BYTE *cipher;
    CK_ULONG cipher_len;
    string cipher_b64;

    iv_len = (CK_ULONG) AES_KEYLENGTH/8;
    iv = (CK_BYTE *)malloc(iv_len * sizeof(CK_BYTE));
    CK_MECHANISM mechanism = {
      CKM_AES_CBC_PAD, iv, iv_len
    };
    Base64::Encode(string((char *)iv, iv_len), &iv_b64);

    rv = p11->C_EncryptInit(hSession, &mechanism, hKey);
    if (rv == CKR_OK) {
        rv = p11->C_Encrypt(hSession, (CK_BYTE_PTR) plain.c_str(), plain.length(), NULL, &cipher_len);
        if (rv == CKR_OK) {
            cipher = (CK_BYTE *)malloc(cipher_len * sizeof(CK_BYTE));
            rv = p11->C_Encrypt(hSession, (CK_BYTE_PTR) plain.c_str(), plain.length(), cipher, &cipher_len);
        }
    }
    if (rv != CKR_OK) {
        cout << "Encryption failed. Error code = " << rv << endl;
        return string();
    }
    Base64::Encode(string((char *)cipher, cipher_len), &cipher_b64);
    return iv_b64 + ":" + cipher_b64;
}

string decrypt(CK_FUNCTION_LIST_PTR p11, CK_SESSION_HANDLE hSession, CK_OBJECT_HANDLE hKey, string iv_cipher) {
    CK_RV rv;
    CK_BYTE *plain;
    CK_ULONG plain_len;
    string iv;
    string cipher;

    Base64::Decode(iv_cipher.substr(0, iv_cipher.find(':')), &iv);
    CK_MECHANISM mechanism = {
      CKM_AES_CBC_PAD, (CK_BYTE_PTR) iv.c_str(), iv.length()
    };

    rv = p11->C_DecryptInit(hSession, &mechanism, hKey);
    if (rv == CKR_OK) {
	      Base64::Decode(iv_cipher.substr(iv_cipher.find(':')+1, iv_cipher.length() - iv_cipher.find(':') + 1), &cipher);
        rv = p11->C_Decrypt(hSession, (CK_BYTE_PTR) cipher.c_str(), cipher.length(), NULL, &plain_len);
        if (rv == CKR_OK) {
			      plain = (CK_BYTE *)malloc(plain_len * sizeof(CK_BYTE));
            rv = p11->C_Decrypt(hSession, (CK_BYTE_PTR) cipher.c_str(), cipher.length(), plain, &plain_len);
        }
    }
    if (rv != CKR_OK) {
        cout << "Decryption failed. Error code = " << rv << endl;
				return string();
    }
    return string((char*)plain, plain_len);
}

CK_RV delete_key(CK_FUNCTION_LIST_PTR p11, CK_SESSION_HANDLE hSession, CK_OBJECT_HANDLE hKey) {
	return p11->C_DestroyObject(hSession, hKey);
}

CK_RV logout(CK_FUNCTION_LIST_PTR p11, CK_SESSION_HANDLE hSession) {
	return p11->C_Logout(hSession);
}

int main(int argc, char *argv[]) {
    CK_FUNCTION_LIST_PTR p11;
    CK_SESSION_HANDLE hSession;
    CK_OBJECT_HANDLE hKey;
    string cipher;
    string decrypted_data;
    string pin;
		
    if (argc == 2) {
        pin = string(argv[1]);
    } else {
        cout << "Enter API Key: ";
        cin >> pin;
    }

    cout << endl;

    if (!(p11 = initialize())) return 1;
    if (login(p11, &hSession, pin) != CKR_OK) return 1;
    if ((hKey = generate_key(p11, hSession, AES_KEYLENGTH/8)) == CK_INVALID_HANDLE) return 1;
    cout << "Encrypting string: \"" << str_to_encrypt << "\"" << endl;
    if ((cipher=encrypt(p11, hSession, hKey, str_to_encrypt)) == string()) {
			delete_key(p11, hSession, hKey);
			return 1;
		}
    cout << "Encrypted string: \"" << cipher << "\"" << endl;
    if ((decrypted_data=decrypt(p11, hSession, hKey, cipher)) == string()) {
			delete_key(p11, hSession, hKey);	
			return 1;
		}
    cout << "Decrypted string: \"" << decrypted_data << "\"" << endl;
	if (delete_key(p11, hSession, hKey) != CKR_OK) return 1;
	if (logout(p11, hSession) != CKR_OK) return 1;
    return 0;
}
